//-----------------------------------------------------------------------------
// NumberVar.c
//   Defines the routines for handling numeric variables.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Number type return values
//-----------------------------------------------------------------------------
static const int gc_NumberAsUnknown = 0;
static const int gc_NumberAsFloat = 1;
static const int gc_NumberAsInteger = 2;
static const int gc_NumberAsString = 3;
static const int gc_NumberAsLongInteger = 4;
static const char gc_NumberFormat[63] =
        "999999999999999999999999999999999999999999999999999999999999999";

//-----------------------------------------------------------------------------
// Number type
//-----------------------------------------------------------------------------
typedef struct {
    Variable_HEAD
    OCINumber *data;
    int returnType;
} udt_NumberVar;


//-----------------------------------------------------------------------------
// Declaration of number variable functions.
//-----------------------------------------------------------------------------
static int NumberVar_Initialize(udt_NumberVar*);
static int NumberVar_PreDefine(udt_NumberVar*, OCIParam*);
static int NumberVar_SetValue(udt_NumberVar*, unsigned, PyObject*);
static PyObject *NumberVar_GetValue(udt_NumberVar*, unsigned);


//-----------------------------------------------------------------------------
// Python type declaration
//-----------------------------------------------------------------------------
static PyTypeObject g_NumberVarType = {
    PyObject_HEAD_INIT(NULL)
    0,					// ob_size
    "cx_Oracle.NUMBER",			// tp_name
    sizeof(udt_NumberVar),		// tp_basicsize
    0,					// tp_itemsize
    (destructor) Variable_Free,		// tp_dealloc
    0,					// tp_print
    0,					// tp_getattr
    0,					// tp_setattr
    0,					// tp_compare
    0,					// tp_repr
    0,					// tp_as_number
    0,					// tp_as_sequence
    0,					// tp_as_mapping
    0,					// tp_hash
    0,					// tp_call
    0,					// tp_str
    (getattrofunc) Variable_GetAttr,	// tp_getattro
    0,					// tp_setattro
    0,					// tp_as_buffer
    Py_TPFLAGS_DEFAULT,			// tp_flags
    0					// tp_doc
};


//-----------------------------------------------------------------------------
// variable type declarations
//-----------------------------------------------------------------------------
static udt_VariableType vt_Number = {
    (InitializeProc) NumberVar_Initialize,
    (FinalizeProc) NULL,
    (PreDefineProc) NumberVar_PreDefine,
    (PostDefineProc) NULL,
    (SetValueProc) NumberVar_SetValue,
    (GetValueProc) NumberVar_GetValue,
    &g_NumberVarType,			// Python type
    SQLT_VNU,				// Oracle type
    sizeof(OCINumber),			// element length
    0,					// is variable length
    1,					// can be copied
    1					// can be in array
};


//-----------------------------------------------------------------------------
// NumberVar_Initialize()
//   Initialize the variable.
//-----------------------------------------------------------------------------
static int NumberVar_Initialize(
    udt_NumberVar *var)			// variable to initialize
{
    var->returnType = gc_NumberAsUnknown;
    return 0;
}


//-----------------------------------------------------------------------------
// NumberVar_PreDefine()
//   Set the type of value (integer, float or string) that will be returned
// when values are fetched from this variable.
//-----------------------------------------------------------------------------
static int NumberVar_PreDefine(
    udt_NumberVar *var,			// variable to initialize
    OCIParam *param)			// parameter handle
{
    sb2 precision;
    sword status;
    sb1 scale;

    // return a string if the global flag is set
    if (PyDict_GetItemString(g_ModuleDictionary, "OPT_NumbersAsStrings"))
        var->returnType = gc_NumberAsString;

    // otherwise, check to see if the number can fit inside an integer by
    // looking at the precision and scale
    else {
        scale = precision = 0;
        status = OCIAttrGet(param, OCI_HTYPE_DESCRIBE, (dvoid*) &scale, 0,
                OCI_ATTR_SCALE, var->environment->errorHandle);
        if (Environment_CheckForError(var->environment, status,
                "NumberVar_PreDefine(): scale") < 0)
            return -1;
        status = OCIAttrGet(param, OCI_HTYPE_DESCRIBE, (dvoid*) &precision, 0,
                OCI_ATTR_PRECISION, var->environment->errorHandle);
        if (Environment_CheckForError(var->environment, status,
                "NumberVar_PreDefine(): precision") < 0)
            return -1;
        if (scale == 0 && precision > 0 && precision < 10)
            var->returnType = gc_NumberAsInteger;
        else if (scale == 0 || (scale == -127 && precision == 0))
            var->returnType = gc_NumberAsLongInteger;
        else var->returnType = gc_NumberAsFloat;
    }

    return 0;
}


//-----------------------------------------------------------------------------
// NumberVar_SetValue()
//   Set the value of the variable.
//-----------------------------------------------------------------------------
static int NumberVar_SetValue(
    udt_NumberVar *var,			// variable to set value for
    unsigned pos,			// array position to set
    PyObject *value)			// value to set
{
    PyObject *textValue;
    double doubleValue;
    int integerValue;
    sword status;

    // make sure a number is being bound
    if (!PyInt_Check(value) && !PyFloat_Check(value)
            && !PyLong_Check(value)) {
        PyErr_SetString(PyExc_TypeError, "expecting numeric data");
        return -1;
    }

    // otherwise, copy the number into the array
    if (PyInt_Check(value)) {
        integerValue = PyInt_AS_LONG(value);
        status = OCINumberFromInt(var->environment->errorHandle, &integerValue,
                sizeof(int), OCI_NUMBER_SIGNED, &var->data[pos]);
        if (var->returnType == gc_NumberAsUnknown)
            var->returnType = gc_NumberAsInteger;
    } else if (PyFloat_Check(value)) {
        doubleValue = PyFloat_AS_DOUBLE(value);
        status = OCINumberFromReal(var->environment->errorHandle, &doubleValue,
                sizeof(double), &var->data[pos]);
        var->returnType = gc_NumberAsFloat;
    } else {
        textValue = PyObject_Str(value);
        if (!textValue)
            return -1;
        status = OCINumberFromText(var->environment->errorHandle,
                PyString_AS_STRING(textValue), PyString_GET_SIZE(textValue),
                gc_NumberFormat, sizeof(gc_NumberFormat), NULL, 0,
                &var->data[pos]);
        Py_DECREF(textValue);
    }
    if (Environment_CheckForError(var->environment, status,
            "NumberVar_SetValue()") < 0)
        return -1;

    return 0;
}


//-----------------------------------------------------------------------------
// NumberVar_GetValue()
//   Returns the value stored at the given array position.
//-----------------------------------------------------------------------------
static PyObject *NumberVar_GetValue(
    udt_NumberVar *var,			// variable to determine value for
    unsigned pos)			// array position
{
    char stringValue[200];
    double doubleValue;
    int integerValue;
    ub4 stringLength;
    PyObject *result;
    sword status;

    if (var->returnType == gc_NumberAsInteger) {
        status = OCINumberToInt(var->environment->errorHandle, &var->data[pos],
                sizeof(int), OCI_NUMBER_SIGNED, (dvoid*) &integerValue);
        if (Environment_CheckForError(var->environment, status,
                "NumberVar_GetValue(): as integer") < 0)
            return NULL;
        return PyInt_FromLong(integerValue);
    }

    if (var->returnType == gc_NumberAsString ||
            var->returnType == gc_NumberAsLongInteger) {
        stringLength = sizeof(stringValue);
        status = OCINumberToText(var->environment->errorHandle,
                &var->data[pos], (unsigned char*) "TM9", 3, NULL, 0,
                &stringLength, (unsigned char*) stringValue);
        if (Environment_CheckForError(var->environment, status,
                "NumberVar_GetValue(): as string") < 0)
            return NULL;
        if (var->returnType == gc_NumberAsString)
            return PyString_FromStringAndSize(stringValue, stringLength);
        result = PyInt_FromString(stringValue, NULL, 10);
        if (result || !PyErr_ExceptionMatches(PyExc_ValueError))
            return result;
        PyErr_Clear();
        result = PyLong_FromString(stringValue, NULL, 10);
        if (result || !PyErr_ExceptionMatches(PyExc_ValueError))
            return result;
        PyErr_Clear();
    }

    status = OCINumberToReal(var->environment->errorHandle,
            &var->data[pos], sizeof(double), (dvoid*) &doubleValue);
    if (Environment_CheckForError(var->environment, status,
            "NumberVar_GetValue(): as float") < 0)
        return NULL;
    return PyFloat_FromDouble(doubleValue);
}

